import math
import moderngl
import pyrr
import numpy
from PyQt5 import QtOpenGL, QtGui, QtCore
import util_moderngl_qt.view_navigation3

class QtGLWidget_Viewer3(QtOpenGL.QGLWidget):

    def __init__(self, list_drawer=[], parent=None):
        self.parent = parent
        fmt = QtOpenGL.QGLFormat()
        fmt.setVersion(3, 3)
        fmt.setProfile(QtOpenGL.QGLFormat.CoreProfile)
        fmt.setSampleBuffers(True)
        super(QtGLWidget_Viewer3, self).__init__(fmt, None)
        #
        self.nav = util_moderngl_qt.view_navigation3.ViewNavigation3()
        self.resize(640, 480)
        self.setWindowTitle('Mesh Viewer')
        self.list_drawer = list_drawer
        self.mousePressCallBack = []
        self.mouseMoveCallBack = []

    def initializeGL(self):
        self.ctx = moderngl.create_context()
        self.ctx.enable(moderngl.DEPTH_TEST)
        for drawer in self.list_drawer:
            drawer.init_gl(self.ctx)

    def paintGL(self):
        self.ctx.clear(1.0, 0.8, 1.0)
        self.ctx.polygon_offset = 1.1, 4.0
        proj = self.nav.projection_matrix()
        modelview = self.nav.modelview_matrix()
        zinv = pyrr.Matrix44(value=(1,0,0,0, 0,1,0,0, 0,0,-1,0, 0,0,0,1),dtype=numpy.float32)
        for drawer in self.list_drawer:
            drawer.paint_gl(zinv*proj*modelview)

    def resizeGL(self, width, height):
        width = max(2, width)
        height = max(2, height)
        self.ctx.viewport = (0, 0, width, height)
        self.nav.win_height = self.height()
        self.nav.win_width = self.width()

    def mousePressEvent(self, event):
        self.nav.update_cursor_position(event.pos().x(), event.pos().y())
        if event.buttons() & QtCore.Qt.LeftButton:
            self.nav.btn_left = True
        for cb in self.mousePressCallBack:
            cb(event)

    def mouseReleaseEvent(self, event):
        if event.buttons() & QtCore.Qt.LeftButton:
            self.nav.btn_left = False

    def mouseMoveEvent(self, event):
        self.nav.update_cursor_position(event.pos().x(), event.pos().y())
        if event.buttons() & QtCore.Qt.LeftButton:
            if event.modifiers() & QtCore.Qt.KeyboardModifier.ShiftModifier:
                self.nav.camera_translation()
                self.update()
            if event.modifiers() & QtCore.Qt.KeyboardModifier.AltModifier:
                self.nav.camera_rotation()
                self.update()
        for cb in self.mouseMoveCallBack:
            cb(event)

    def wheelEvent(self, event: QtGui.QWheelEvent) -> None:
        dy = event.pixelDelta().y()
        self.nav.scale *= math.pow(1.01, dy)
        self.update()



from pyrr import Matrix44
import moderngl


class DrawerTransformMulti:

    def __init__(self, drawer):
        self.list_transform = []
        self.drawer = drawer

    def init_gl(self, ctx: moderngl.Context):
        self.drawer.init_gl(ctx)

    def paint_gl(self, mvp: Matrix44):
        for transform in self.list_transform:
            self.drawer.paint_gl(mvp * transform)



from pyrr import Matrix44
import numpy
import moderngl


class DrawerTransform:

    def __init__(self, drawer):
        self.transform = Matrix44.identity(dtype=numpy.float32)
        self.is_visible = True
        self.drawer = drawer

    def init_gl(self, ctx: moderngl.Context):
        self.drawer.init_gl(ctx)

    def paint_gl(self, mvp: Matrix44):
        if not self.is_visible:
            return
        self.drawer.paint_gl(mvp * self.transform)



from pyrr import Matrix44
import numpy
from moderngl.vertex_array import TRIANGLES

class DrawerMesPosColor:

    def __init__(self, V:bytes, C:bytes, F:bytes):
        self.V = V
        self.C = C
        self.F = F

    def init_gl(self, ctx):
        self.prog = ctx.program(
            vertex_shader='''
                #version 330
                uniform mat4 Mvp;
                in vec3 in_position;
                in vec3 in_color;
                out vec3 v_vert;
                out vec3 v_color;
                void main() {
                    v_vert = in_position;
                    v_color = in_color;
                    gl_Position = Mvp * vec4(in_position, 1.0);
                }
            ''',
            fragment_shader='''
                #version 330
                in vec3 v_vert;
                in vec3 v_color;
                out vec4 f_color;
                void main() {
                    f_color = vec4(v_color, 1.0);
                }
            '''
        )
        self.mvp = self.prog['Mvp']

        vao_content = [
            (ctx.buffer(self.V), '3f', 'in_position'),
            (ctx.buffer(self.C), '3f', 'in_color')
        ]
        index_buffer = ctx.buffer(self.F)

        self.vao = ctx.vertex_array(
            self.prog, vao_content, index_buffer, 4)

    def paint_gl(self, mvp: Matrix44):
        self.mvp.value = tuple(mvp.flatten())
        self.vao.render(mode=TRIANGLES)


import typing
from pyrr import Matrix44
import numpy
import moderngl

class ElementInfo:

    def __init__(self, index, mode, color):
        self.vao = None
        self.index = index
        self.mode = mode
        self.color = color

class DrawerMesPosTex:

    def __init__(self,
                 list_elem2vtx: typing.List[ElementInfo],
                 vtx2xyz: numpy.ndarray,
                 vtx2uv: numpy.ndarray):
        self.vtx2xyz = vtx2xyz
        self.vtx2uv = vtx2uv
        self.list_elem2vtx = list_elem2vtx
        self.vao_content = None

    def init_gl(self, ctx: moderngl.Context):
        self.prog = ctx.program(
            vertex_shader='''
                #version 330
                uniform mat4 Mvp;
                in vec2 in_uv;
                in vec3 in_xyz;
                out vec2 texPrj;
                void main() {
                    gl_Position = Mvp * vec4(in_xyz, 1.0);
                    texPrj = in_uv;
                }
            ''',
            fragment_shader='''
                #version 330
                uniform sampler2D myTextureSampler;
                uniform vec3 color;
                uniform bool is_texture;
                out vec4 f_color;
                in vec2 texPrj;
                void main() {
                    if( is_texture ) {                 
                        f_color = texture(myTextureSampler,texPrj);
                    }
                    else { 
                        f_color = vec4(color, 1.0);
                    }
                }
            '''
        )
        self.uniform_mvp = self.prog['Mvp']
        self.uniform_texture_location = self.prog['myTextureSampler']
        self.uniform_is_texture = self.prog['is_texture']
        self.uniform_color = self.prog['color']

        self.vao_content = [
            (ctx.buffer(self.vtx2xyz.tobytes()), '3f', 'in_xyz'),
            (ctx.buffer(self.vtx2uv.tobytes()), '2f', 'in_uv')
        ]
        del self.vtx2xyz
        del self.vtx2uv
        for el in self.list_elem2vtx:
            index_buffer = ctx.buffer(el.index.tobytes())
            el.vao = ctx.vertex_array(
                self.prog, self.vao_content, index_buffer, 4
            )
            del el.index

    def update_position(self, vtx2xyz: numpy.ndarray):
        if self.vao_content != None:
            vbo = self.vao_content[0][0]
            vbo.write(vtx2xyz.tobytes())

    def paint_gl(self, mvp: Matrix44, texture_location: int):
        self.uniform_mvp.value = tuple(mvp.flatten())
        self.uniform_texture_location.value = texture_location
        for el in self.list_elem2vtx:
            if el.color is None:
                self.uniform_is_texture.value = True
            else:
                self.uniform_is_texture.value = False
            el.vao.render(mode=el.mode)


import math, numpy

import pyrr
from pyrr import Quaternion, Vector3, Matrix44


class ViewNavigation3:
    def __init__(self, view_height=1.0):
        self.view_height = view_height
        self.scale = 1.
        self.depth_ratio = 10.
        #
        self.translation = Vector3((0., 0., 0.), dtype=numpy.float64)
        self.quat = Quaternion((0., 0., 0., 1.), dtype=numpy.float64)
        #
        self.cursor_x = 0.
        self.cursor_y = 0.
        self.cursor_dx = 0.
        self.cursor_dy = 0.
        self.win_height = 480
        self.win_width = 640

    def update_cursor_position(self, x, y):
        fw = self.win_width
        fh = self.win_height
        x0 = self.cursor_x
        y0 = self.cursor_y
        self.cursor_x = (2.0 * x - fw) / fw
        self.cursor_y = (fh - 2.0 * y) / fh
        self.cursor_dx = self.cursor_x - x0
        self.cursor_dy = self.cursor_y - y0

    def projection_matrix(self) -> pyrr.Matrix44:
        asp = float(self.win_width) / float(self.win_height)
        mp = Matrix44(
            (1. / (self.view_height * asp), 0., 0., 0.,
             0., 1. / self.view_height, 0., 0.,
             0., 0., 1. / (self.view_height * self.depth_ratio), 0.,
             0., 0., 0., 1.))
        ms = Matrix44.from_scale((self.scale, self.scale, self.scale))
        return ms * mp

    def modelview_matrix(self) -> pyrr.Matrix44:
        mt = Matrix44.from_translation(self.translation)
        mr = Matrix44.from_quaternion(self.quat)
        return mt * mr

    def camera_rotation(self):
        dx = self.cursor_dx
        dy = self.cursor_dy
        a = math.sqrt(dx * dx + dy * dy)
        if a == 0.0:
            return
        dq = Quaternion.from_axis((dy, -dx, 0.))
        dq.normalize()
        self.quat = self.quat * dq

    def camera_translation(self):
        mp = self.projection_matrix()
        sx = (mp[3, 3] - mp[0, 3]) / mp[0, 0]
        sy = (mp[3, 3] - mp[1, 3]) / mp[1, 1]
        self.translation[0] += sx * self.cursor_dx
        self.translation[1] += sy * self.cursor_dy

    def picking_ray(self) -> (pyrr.Vector4, pyrr.Vector4):
        mmv = self.modelview_matrix()
        mp = self.projection_matrix()
        mmvpi = (mp * mmv).inverse
        q0 = mmvpi * pyrr.Vector4([self.cursor_x, self.cursor_y, +1., 1.])
        q1 = mmvpi * pyrr.Vector4([self.cursor_x, self.cursor_y, -1., 1.])
        return q0, q1 - q0



import typing
from pyrr import Matrix44
import numpy
import moderngl

class ElementInfo:

    def __init__(self, index, mode, color):
        self.vao = None
        self.index = index
        self.mode = mode
        self.color = color

class DrawerMesPos:

    def __init__(self, V:numpy.ndarray, element: typing.List[ElementInfo]):
        self.V = V
        self.element = element
        self.vao_content = None

    def init_gl(self, ctx: moderngl.Context):
        self.prog = ctx.program(
            vertex_shader='''
                #version 330
                uniform mat4 Mvp;
                in vec3 in_position;
                void main() {
                    gl_Position = Mvp * vec4(in_position, 1.0);
                }
            ''',
            fragment_shader='''
                #version 330
                uniform vec3 color;                
                out vec4 f_color;
                void main() {
                    f_color = vec4(color, 1.0);
                }
            '''
        )
        self.uniform_mvp = self.prog['Mvp']
        self.uniform_color = self.prog['color']

        self.vao_content = [
            (ctx.buffer(self.V.tobytes()), '3f', 'in_position'),
        ]
        del self.V
        for el in self.element:
            index_buffer = ctx.buffer(el.index.tobytes())
            el.vao = ctx.vertex_array(
                self.prog, self.vao_content, index_buffer, 4
            )
            del el.index

    def update_position(self, V: numpy.ndarray):
        if self.vao_content != None:
            vbo = self.vao_content[0][0]
            vbo.write(V.tobytes())

    def paint_gl(self, mvp: Matrix44):
        self.uniform_mvp.value = tuple(mvp.flatten())
        for el in self.element:
            self.uniform_color.value = el.color
            el.vao.render(mode=el.mode)


import math
import pyrr
import moderngl
import numpy
from PyQt5 import QtOpenGL, QtWidgets, QtCore, QtGui

import util_moderngl_qt.qtglwidget_viewer3

class QtGLWidget_Viewer3_Texture(QtOpenGL.QGLWidget):

    def __init__(self, drawer, img: numpy.ndarray, parent=None):
        self.parent = parent
        fmt = QtOpenGL.QGLFormat()
        fmt.setVersion(3, 3)
        fmt.setProfile(QtOpenGL.QGLFormat.CoreProfile)
        fmt.setSampleBuffers(True)
        super(QtGLWidget_Viewer3_Texture, self).__init__(fmt, None)
        #
        self.nav = util_moderngl_qt.view_navigation3.ViewNavigation3()
        self.resize(640, 480)
        self.setWindowTitle('Mesh Viewer')
        self.drawer = drawer
        self.img = img
        self.mousePressCallBack = []
        self.mouseMoveCallBack = []

    def initializeGL(self):
        self.ctx = moderngl.create_context()
        self.drawer.init_gl(self.ctx)
        img2 = numpy.flip(self.img, axis=0)  # flip upside down
        self.texture = self.ctx.texture((img2.shape[0], img2.shape[1]), img2.shape[2], img2.tobytes())
        del self.img

    def paintGL(self):
        self.ctx.clear(0.9, 0.8, 1.0)
        self.ctx.enable(moderngl.DEPTH_TEST)
        proj = self.nav.projection_matrix()
        modelview = self.nav.modelview_matrix()
        zinv = pyrr.Matrix44(value=(1,0,0,0, 0,1,0,0, 0,0,-1,0, 0,0,0,1),dtype=numpy.float32)
        mvp = zinv * proj * modelview
        self.texture.use(location=0)
        self.drawer.paint_gl(mvp,0)

    def resizeGL(self, width, height):
        width = max(2, width)
        height = max(2, height)
        self.ctx.viewport = (0, 0, width, height)

    def mousePressEvent(self, event):
        self.nav.update_cursor_position(event.pos().x(), event.pos().y())
        if event.buttons() & QtCore.Qt.LeftButton:
            self.nav.btn_left = True
        for cb in self.mousePressCallBack:
            cb(event)

    def mouseReleaseEvent(self, event):
        if event.buttons() & QtCore.Qt.LeftButton:
            self.nav.btn_left = False

    def mouseMoveEvent(self, event):
        self.nav.update_cursor_position(event.pos().x(), event.pos().y())
        if event.buttons() & QtCore.Qt.LeftButton:
            if event.modifiers() & QtCore.Qt.KeyboardModifier.ShiftModifier:
                self.nav.camera_translation()
                self.update()
            if event.modifiers() & QtCore.Qt.KeyboardModifier.AltModifier:
                self.nav.camera_rotation()
                self.update()
        for cb in self.mouseMoveCallBack:
            cb(event)

    def wheelEvent(self, event: QtGui.QWheelEvent) -> None:
        dy = event.pixelDelta().y()
        self.nav.scale *= math.pow(1.01, dy)
        self.update()


from setuptools import setup

def _requires_from_file(filename):
    return open(filename).read().splitlines()

setup(
    name="util_moderngl_qt",
    version="0.1.7",
    install_requires=_requires_from_file('requirements.txt')
)